/*
  best_splits.c

  Implements the helper routines to support randomization of split 
  choices in C4.5. This code and algorithm are based on 

  "An Experimental Comparison of Three Methods for Constructing 
   Ensembles of Decision Trees: Bagging, Boosting, and Randomization"
   Thomas G. Dietterich
   Machine Learning, 40 (2), 139-158.

   Briefly, the most promising splits are maintained when considering
   a split in randomization mode. The number of splits is defined
   elsewhere. Each time a continuous split is evaluated, 
   update_best_splits() should be called to possibly add the split to 
   the list. Each attribute has a separate top-n list, sorted in 
   decreasing order of gain_ratio. Once the average gain is computed
   after all attributes are examined, a final merged list is created
   from each attribute list (the ALL_SPLITS index below).

   Once all splits are evaluated, choose_random_split() should be
   called to pick the split according to a random selection of the
   top most promising splits.

   NOTE: This code is always used; if rand_split_max=1 and/or 
   max_random_level=0 we simply choose the best split (as before).

*/
#include <sys/types.h>
#include <time.h>
#include "defns.h"
#include "types.h"
#include "extern.h"

#define ALL_SPLITS (number_of_attributes)


/* The private global variable to hold all of the splits. The storage
   is malloc'd, but never freed. So sue me */
static SplitList **splits=NULL;

extern float *Gain;
extern float *Info;
extern float *Bar;

extern float BestVal;

/* Private function */
void _consolidate_splits();


/* Initialize and clear split lists */
void initialize_splits()
{
  int i,j;


  /* This is an extra kludge to specify all attributes to choose from, -1 */
  if ( rand_split_max == -1 ) rand_split_max=number_of_attributes;

  /* If the local variable is NULL, do first-time initializations */
  if ( splits == NULL ) {

    /* Allocate list pointers */
    splits=CALLOC(number_of_attributes+1,sizeof(SplitList *));

    /* Allocate list structure */
    for (i=0; i < number_of_attributes+1; i++) {

      /* Only allocate for continuous or the final one */
      if ( i == number_of_attributes || attribute_type[i]==CONTINUOUS) {
         splits[i]=CALLOC(1,sizeof(SplitList));
         /* Allocate the individual pieces */
         splits[i]->attribute=MALLOC(rand_split_max*sizeof(int));
         splits[i]->gain=MALLOC(rand_split_max*sizeof(float));
         splits[i]->info=MALLOC(rand_split_max*sizeof(float));
         splits[i]->gain_ratio=MALLOC(rand_split_max*sizeof(float));
         splits[i]->bar=MALLOC(rand_split_max*sizeof(float));
      }
    }
  }

  /* Always init counters */
  for (i=0; i < number_of_attributes+1; i++) {
    if ( i == number_of_attributes || attribute_type[i] == CONTINUOUS ) {
       splits[i]->number_of_splits=0;
       for (j=0; j < rand_split_max; j++) {
         splits[i]->attribute[j]=-1;
         splits[i]->gain[j]=-Epsilon;
         splits[i]->gain_ratio[j]=-Epsilon;
         splits[i]->info[j]=-Epsilon;
         splits[i]->bar[j]=0.0;
       }
    }
  }
  return;
}

/*
  Given a currently considered split, add it to the list of splits,
  maintained in sorted order (by gain_ratio). The global flag indicates
  if this is by attribute or across attributes.
*/
void update_best_splits(int attribute, double gain, double info, 
		    double gain_ratio, double bar, Boolean global)
{
  SplitList *cur_split;
  int i,j;

  /* Original code did not consider negative gain */
  if ( gain <= -Epsilon ) return;


  if ( global ) cur_split=splits[ALL_SPLITS];
  else          cur_split=splits[attribute];

  
  /* Insert this split into cur_split, moving any values down if
     necessary */
  
  /* First, find the correct place to insert */
  for (i=rand_split_max-1; i >= 0; i--) {
    if ( gain_ratio <= cur_split->gain_ratio[i] ) break;
  }
  /* If the example isn't big enough, return */
  i++;
  if ( i >= rand_split_max ) return;

  /* Insert at the end, or don't insert */
  if ( cur_split->number_of_splits < rand_split_max ) 
    cur_split->number_of_splits++;

  /* Else shuffle everything down (and possibly off the list) */
  for (j=rand_split_max-1; j > i ;j--) {
    cur_split->attribute[j]=cur_split->attribute[j-1];
    cur_split->gain[j]=cur_split->gain[j-1];
    cur_split->info[j]=cur_split->info[j-1];
    cur_split->gain_ratio[j]=cur_split->gain_ratio[j-1];
    cur_split->bar[j]=cur_split->bar[j-1];
  }
  
  /* And assign current values to index i */
  cur_split->attribute[i]=attribute;
  cur_split->gain[i]=gain;
  cur_split->info[i]=info;
  cur_split->gain_ratio[i]=gain_ratio;
  cur_split->bar[i]=bar;  

  return;
}


/*
  choose_random_split returns the index of the split in splits, either
  randomly or by default.  
*/
int choose_split(int current_tree_depth, double average_gain, 
				int *best_attribute, float *best_value)
{
  int choice=0;
  SplitList *all_splits=splits[ALL_SPLITS];
  int num_splits;
  int a;

  /*
    We have to wait until we get to this point to calculate the 
    actual ranking of splits across attributes. We do this now
    across attributes and end up with a single array
  */
  _consolidate_splits(average_gain);
  
  /* If we do not randomize beyond max_random_tree_depth and
     we are beyond this point, skip. */
  if ( rand_split_max > 1 && (max_random_tree_depth < 0 ||
			      current_tree_depth <= max_random_tree_depth) ) {

    /* Randomly choose from the top rand_split_max splits */
    if ( rand_split_max < all_splits->number_of_splits )
      num_splits=rand_split_max;
    else
      num_splits=all_splits->number_of_splits;

    /* Choose the lucky split */
    if ( num_splits > 0 )
       choice=lrand48() % num_splits;
  }

  /* In either case, we have the best split (choice). Set the various
     fields accordingly. */

  a=all_splits->attribute[choice];
  *best_attribute=a;
  *best_value=all_splits->gain_ratio[choice];
  if ( a >= 0 && a < number_of_attributes ) {
    Gain[a]=all_splits->gain[choice];
    Info[a]=all_splits->info[choice];
    Bar[a]=all_splits->bar[choice];
  }

  return a;
}



/*
  This is a gotcha I learned from Dietterich's code - the per-attribute
  ranks are correct, but you cannot compare across attributes until the
  average gain has been computed. This is computed as the average gain
  from the "best" split of each attribute. 
*/
void _consolidate_splits(double average_gain)
{
  int a,i;
  float val;

  for (a=0; a < number_of_attributes; a++) {

    /* If we use random subspaces, ignore the appropriate attributes */
    if ( use_random_subspaces && ! use_attribute[a] ) 
      continue;

    /*
      For all but continuous attributes, simply calculate the single
      gain for the attribute.
    */
    if ( attribute_type[a] == DISCRETE ) {
      if ( Gain[a] >= 0 ) {
	val=Worth(Info[a], Gain[a], average_gain);
	update_best_splits(a, Gain[a], Info[a], val, 0.0, True);
      } 
    } else if ( attribute_type[a] == IGNORE ) {
       continue;
    } else {
      SplitList *t=splits[a];
      int n=t->number_of_splits;

      /* Adjust if using random attribute */
      if ( use_random_attribute_split ) n=min(t->number_of_splits,1);

      /* Continuous attribute */
      for (i=0; i < n; i++) {
	val=Worth(t->info[i], t->gain[i], average_gain);
	update_best_splits(a, t->gain[i], t->info[i], val, t->bar[i], True);
      }

    }
  }

  return;
}


