/*************************************************************************/
/*								 	 */
/*    Central tree-forming algorithm incorporating all criteria  	 */
/*    ---------------------------------------------------------	 	 */
/*								 	 */
/*************************************************************************/


#include "defns.h"
#include "types.h"
#include "extern.h"


float evaluate_attributes(int Fp, int Lp, float num_cases);
float evaluate_random_attributes(int Fp, int Lp, float num_cases);
TREE *FormTree(int Fp,int Lp, int current_tree_depth);
int find_majority_class(int Fp, int Lp);


float
	*Weight,	/* Weight[i]  = current fraction of item i */
	**Freq,		/* Freq[x][c] = no. items of class c with outcome x */
	*ValFreq,	/* ValFreq[x]   = no. items with outcome x */
	*class_distribution;	/* ClassFreq[c] = no. items of class c */


float
	*Gain,		/* Gain[a] = info gain by split on att a */
	*Info,		/* Info[a] = potential info of split on att a */
	*Bar,		/* Bar[a]  = best threshold for contin att a */
	*UnknownRate;	/* UnknownRate[a] = current unknown rate for att a */

Boolean
	*Tested,	/* Tested[a] set if att a has already been tested */
	MultiVal;	/* true when all atts have many values */


	/*  External variables initialised here  */

extern float
	*SplitGain,	/* SplitGain[i] = gain with att value of item i as threshold */
	*SplitInfo;	/* SplitInfo[i] = potential info ditto */

extern float
	*Slice1,	/* Slice1[c]    = saved values of Freq[x][c] in subset.c */
	*Slice2;	/* Slice2[c]    = saved values of Freq[y][c] */

extern Set
	**Subset;	/* Subset[a][s] = subset s for att a */

extern short
	*Subsets;	/* Subsets[a] = no. subsets for att a */

extern float *example_weight;

extern Boolean peel_off;
extern Boolean *real_use_attribute;
int nodeid=0;

/*************************************************************************/
/*								 	 */
/*		Allocate space for tree tables			 	 */
/*								 	 */
/*************************************************************************/
int InitialiseTreeData()
{ 
  int a,v;
  
  Tested=(Boolean *)CALLOC(number_of_attributes, sizeof(Boolean));
  
  Gain	= (float *)CALLOC(number_of_attributes, sizeof(float));
  Info	= (float *)CALLOC(number_of_attributes, sizeof(float));
  Bar	= (float *)CALLOC(number_of_attributes, sizeof(float));
  
  Subset = (Set **)CALLOC(number_of_attributes, sizeof(Set *));
  for (a=0; a < number_of_attributes; a++) {
    if ( MaxAttVal[a] ) {
      Subset[a]  = (Set *) calloc(MaxDiscrVal+1, sizeof(Set));
      ForEach(v, 0, MaxAttVal[a]) {
	Subset[a][v] = (Set) malloc((MaxAttVal[a]>>3) + 1);
      }
    }
  }
  Subsets = (short *) calloc(number_of_attributes, sizeof(short));
  
  SplitGain = (float *)CALLOC(number_of_examples, sizeof(float));
  SplitInfo = (float *)CALLOC(number_of_examples, sizeof(float));
  
  Weight = (float *) calloc(number_of_examples, sizeof(float));
 
  Freq  = (float **) calloc(MaxDiscrVal+1, sizeof(float *));
  for (v=0; v <= MaxDiscrVal; v++)
    Freq[v]  = (float *) calloc(number_of_classes, sizeof(float));
  
  ValFreq = (float *) calloc(MaxDiscrVal+1, sizeof(float));
  class_distribution = (float *) calloc(number_of_classes, sizeof(float));
  
  Slice1 = (float *) calloc(number_of_classes+1, sizeof(float));
  Slice2 = (float *) calloc(number_of_classes+1, sizeof(float));
  
  UnknownRate = (float *) calloc(number_of_attributes, sizeof(float));
  
  /*  Check whether all attributes have many discrete values  */
  MultiVal = True;
  if ( ! SUBSET ) {
    for ( a = 0 ; MultiVal && a < number_of_attributes ; a++ ) {
      if ( attribute_type[a] != IGNORE ) {
	MultiVal = MaxAttVal[a] >= 0.3 * number_of_examples;
      }
    }
  }
  return 0;
}



/*************************************************************************/
/*								 	 */
/*		Initialise the weight of each item		 	 */
/*								 	 */
/*************************************************************************/
int InitialiseWeights()
{
  int i;
  
  for (i=0; i < number_of_examples; i++) {
     if ( use_example_weights )
        Weight[i] = example_weight[i];
     else
        Weight[i] = 1.0;
  }

  return 0;
}



/*************************************************************************/
/*								 	 */
/*  Build a decision tree for the cases Fp through Lp:		 	 */
/*								 	 */
/*  - if all cases are of the same class, the tree is a leaf and so	 */
/*      the leaf is returned labelled with this class		 	 */
/*								 	 */
/*  - for each attribute, calculate the potential information provided 	 */
/*	by a test on the attribute (based on the probabilities of each	 */
/*	case having a particular value for the attribute), and the gain	 */
/*	in information that would result from a test on the attribute	 */
/*	(based on the probabilities of each case with a particular	 */
/*	value for the attribute being of a particular class)		 */
/*								 	 */
/*  - on the basis of these figures, and depending on the current	 */
/*	selection criterion, find the best attribute to branch on. 	 */
/*	Note:  this version will not allow a split on an attribute	 */
/*	unless two or more subsets have at least MINOBJS items. 	 */
/*								 	 */
/*  - try branching and test whether better than forming a leaf	 	 */
/*								 	 */
/*************************************************************************/
TREE *FormTree(int Fp,int Lp, int current_tree_depth)
{ 
  int i, Kp, Ep;
  float num_cases, best_class_count, KnownCases;
  float Factor, average_gain, best_gain;
  int best_attribute, best_class;
  TREE *Node;
  int v;
  Boolean PrevAllKnown;
  
  /* Hack, since Lp is not the upper index now, but the number of
     examples in the list */
  Lp--;

  /* Initialise the split list */
  initialize_splits();

  /* Calculate statistics on the current node */
  num_cases = CountItems(Fp, Lp);
  best_class=find_majority_class(Fp,Lp);
  best_class_count=class_distribution[best_class];

  /* The default is to first create a leaf, then change it if possible */
  Node = create_leaf(nodeid++,class_distribution, 
		     best_class, num_cases, num_cases-best_class_count);
  

  /*  If all cases are of the same class or there are not enough
      cases to divide, the tree is a leaf  */
  if ( class_distribution[best_class] == num_cases  || num_cases < 2 * MINOBJS || 
       ( max_tree_depth >= 0 && current_tree_depth >= max_tree_depth ) ) {
    return Node;
  }

  Verbosity(1)
    printf("\n%d items, total weight %.1f\n", Lp - Fp + 1, num_cases);


  /* Evaluate the attributes */
  if ( use_random_forest ) 
    average_gain=evaluate_random_attributes(Fp,Lp,num_cases);
  else 
    average_gain=evaluate_attributes(Fp,Lp,num_cases);


  /* We choose the split, either by best gain or randomly */
  choose_split(current_tree_depth, average_gain, &best_attribute, &best_gain);

  /*  Decide whether to branch or not  */ 
  if ( best_attribute != None ) {
    Verbosity(1) {
      printf("\tbest attribute %s", attribute_name[best_attribute]);
      if ( ! MaxAttVal[best_attribute] ) 
	printf(" cut %.3f", Bar[best_attribute]);
      printf(" inf %.3f gain %.3f val %.3f\n",
	     Info[best_attribute], Gain[best_attribute], best_gain);
    }	

    /*  Build a node of the selected test  */
    if ( attribute_type[best_attribute] == DISCRETE && SUBSET ) {
       if ( MaxAttVal[best_attribute] > 2 ) SubsetTest(Node,best_attribute);
    } else if ( attribute_type[best_attribute] == DISCRETE ) {
	DiscreteTest(Node, best_attribute);
    } else { 
      /*  Continuous attribute  */
      ContinTest(Node, best_attribute);
    } 

    /*  Remove unknown attribute values  */
    PrevAllKnown = AllKnown;

    Kp = group_unknown_values(Fp, Lp, Node) + 1;
    if ( Kp != Fp ) AllKnown = False;
    KnownCases = num_cases - CountItems(Fp, Kp-1);
    UnknownRate[best_attribute] = (num_cases - KnownCases) / (num_cases + 0.001);
    
    Verbosity(1) {
      if ( UnknownRate[best_attribute] > 0 ) {
	printf("\tunknown rate for %s = %.3f\n",
	       attribute_name[best_attribute], UnknownRate[best_attribute]);
      }
    }
    
    /*  Recursive divide and conquer  */
    Tested[best_attribute]=True;
    Ep = Kp - 1;
    Node->Errors = 0;
    
    for (v=1; v <= Node->num_branches; v++) {
      Ep = Group(v, Kp, Lp, Node);
      
      if ( Kp <= Ep ) {
	Factor = CountItems(Kp, Ep) / KnownCases;
	
	for (i=Fp; i <= Kp-1; i++) 
	  Weight[i] *= Factor;
	
	Node->branch[v] = FormTree(Fp, Ep+1,current_tree_depth+1);
	Node->Errors += Node->branch[v]->Errors;
	
	group_unknown_values(Fp, Ep, Node);
	for (i=Fp; i <= Kp-1; i++) {
	  Weight[i] /= Factor;
	}
      } else {
        /* This should be the case in which no example has the
           given attribute along the given branch, a default leaf */
	Node->branch[v]=create_leaf(nodeid++,Node->class_distribution, best_class, 0.0, 0.0);
      }
    }
    Tested[best_attribute]=False;
    if (peel_off ) use_attribute[best_attribute]=False;
    AllKnown = PrevAllKnown;
    
    /*  See whether we would have been no worse off with a leaf  */
    if ( collapse_tree && Node->Errors >= num_cases - best_class_count - Epsilon ) {
      Verbosity(1)
	printf("Collapse tree for %d items to leaf %s\n",
	       Lp - Fp + 1, class_name[best_class]);
      Node->node_type = LEAF_NODE;
      Node->num_branches=0;
    } 
  } else { 
    Verbosity(1)
      printf("\tno sensible splits  %.1f/%.1f\n",
	     num_cases, num_cases - best_class_count);
  } 
  
  return Node; 
} 




/*
  evaluate all attributes according to information gain (or gain ratio).
  The individual Eval routine updates the best split list
*/
float evaluate_attributes(int Fp, int Lp, float num_cases)
{
  int a,possible_attributes=0;
  float average_gain=0;

  /*  For each available attribute, find the information and gain  */
  for (a=0; a < number_of_attributes; a++) {
    Gain[a] = -Epsilon;

    /* Possibly exclude this attribute */
    if ( attribute_type[a] == IGNORE ) continue;
    if ( use_random_subspaces && ! use_attribute[a] )  continue;
    if ( peel_off && ! real_use_attribute[a] ) continue; 
    if ( attribute_type[a] == DISCRETE || attribute_type[a] == BINARY ) {
      /*  discrete valued attribute  */
      if ( SUBSET && MaxAttVal[a] > 2 ) {
	EvalSubset(a, Fp, Lp, num_cases);
      } else if ( ! Tested[a] ) {
	EvalDiscreteAtt(a, Fp, Lp, num_cases);
      }
    } else { 
      /*  continuous attribute  */
      EvalContinuousAtt(a, Fp, Lp);
    } 

    /*  Update average gain, excluding attributes with very many values  */
    if ( Gain[a] > -Epsilon &&
	 ( MultiVal || MaxAttVal[a] < 0.3 * number_of_examples ) ) {
      possible_attributes++;
      average_gain += Gain[a];
    }
  } 


  /* Calculate the average gain, based on number of potential attribute splits */
  average_gain  = ( possible_attributes? average_gain / possible_attributes : 1E6 );
  if ( VERBOSITY >= 2 && average_gain < 1E6 )
    printf("\taverage gain %.3f\n",average_gain);


  return average_gain;
}



/*
  evaluate a random subset of attributes according to information
  gain (or gain ratio). If no gain > -epsilon the process continues 
  until all attributes are exhausted or we find one.
  The individual Eval routines updates the best split list

  Part of this code was adapted from the weka random forest code.

  A variable guide:
    k is the number of attributes we want to select from
    attrs_used is the number we have actually looked at
    a is the currently considered attribute
*/
float evaluate_random_attributes(int Fp, int Lp, float num_cases)
{
  int a,possible_attributes=0,k, attrs_used=0;
  float average_gain=0;
  Boolean found_gain=False;
  Boolean *used;

  used=CALLOC(number_of_attributes, sizeof(Boolean));

  /* There is a heuristic in Breiman's code of sqrt(attr). There is also a log. */
  k=min(number_of_random_attributes, number_of_attributes);
  if ( k == 0 ) k=(int)ceil(sqrt(number_of_attributes));
  if ( k == -1 ) k=(int)log(number_of_attributes)+1;

  if ( VERBOSITY >= 1 ) printf("RandomForest: Evaluating over %d attribute(s)\n",k);
  
  /* Important! Initialize all gain's to -epsilon in case we don't look at them */
  for (a=0; a < number_of_attributes; a++ )
    Gain[a] = -Epsilon;
  

  /* attrs_used keeps track of the total number of attributes (including
     ignored attributes and negative gain attributes)
  */
  while ( attrs_used < number_of_attributes && (!found_gain || k > 0 ) ) {
    attrs_used++;
    k--;

    do { a=lrand48() % number_of_attributes; } while (used[a]);
    used[a]=True;


    /* Possibly exclude this attribute */
    if ( attribute_type[a] == IGNORE ) continue;

 
    if ( attribute_type[a] == DISCRETE || attribute_type[a] == BINARY ) {
      /*  discrete valued attribute  */
      if ( SUBSET && MaxAttVal[a] > 2 ) {
	EvalSubset(a, Fp, Lp, num_cases);
      } else if ( ! Tested[a] ) {
	EvalDiscreteAtt(a, Fp, Lp, num_cases);
      }
    } else { 
      /*  continuous attribute  */
      EvalContinuousAtt(a, Fp, Lp);
    } 

    /* If some gain was found for this attribute */
    if ( Gain[a] > -Epsilon ) {
      found_gain=True;
      /*  Update average gain, excluding attributes with very many values  */
      if ( MultiVal || MaxAttVal[a] < 0.3 * number_of_examples ) {
	possible_attributes++;
	average_gain += Gain[a];
      }
    } 

    /* End of attribute evaluation */
  }

  /* Calculate the average gain, based on number of potential attribute splits */
  average_gain  = ( possible_attributes? average_gain / possible_attributes : 1E6 );
  if ( VERBOSITY >= 2 && average_gain < 1E6 )
    printf("\taverage gain %.3f\n",average_gain);

  free(used);

  if ( VERBOSITY >= 1 ) printf("RandomForest: Evaluated %d total attributes\n",attrs_used);

  return average_gain;
}


/*
  Count up the number of examples of each class that exist at the current node,
  and return the majority class.
*/
int find_majority_class(int Fp, int Lp)
{
  int i,c;
  int majority_class=0;

  /*  Generate the class frequency distribution  */
  for (c=0; c < number_of_classes; c++) 
    class_distribution[c] = 0;
  for (i=Fp; i <= Lp; i++) 
    class_distribution[ Class(Item[i]) ] += Weight[i];
  
  /*  Find the most frequent class  */
  for (c=0; c < number_of_classes; c++) {
    if ( class_distribution[c] > class_distribution[majority_class] ) 
      majority_class = c;
  }

  return majority_class;
}







/*************************************************************************/
/*								 	 */
/*  Group together the items corresponding to branch V of a test 	 */
/*  and return the index of the last such			 	 */
/*								 	 */
/*************************************************************************/
int Group(int V, int Fp,int Lp,TREE *t)
{
  int i,a;
  Set SS;
    
  a = t->tested_attribute;
  
  /*  Group items on the value of attribute a, and depending
      on the type of branch  */
  
  switch ( t->node_type ) {
  case DISCRETE_NODE:
    for (i=Fp; i <= Lp; i++) {
      if ( Item[i][a].txt_index == V ) 
	Swap(Fp++, i);
    }
    break;
  case CONTINUOUS_NODE:
    if ( V == 1 ) {
      for (i=Fp; i <= Lp; i++) {
	if ( Item[i][a].real_value <= t->Cut )
	  Swap(Fp++,i);
      }
    } else {
      for (i=Fp; i <= Lp; i++) {
	if ( Item[i][a].real_value > t->Cut )
	  Swap(Fp++,i);
      }
    }

	    //      if (( Item[i][a].real_value <= t->Cut ) ==  ( V == 1) ) 
	    //Swap(Fp++, i);
    break;
  case SUBSET_NODE:
    SS = t->Subset[V];
    for (i=Fp; i <= Lp; i++) {
      if ( In(DVal(Item[i], a), SS) ) Swap(Fp++, i);
    }
    break;
  }
  
  return Fp - 1;
}



/* Group the unknown values together */
int group_unknown_values(int Fp, int Lp, TREE *t)
{
  int i,a;

  a=t->tested_attribute;

  /*  Group together unknown values  */
  switch ( t->node_type ) {
  case DISCRETE_NODE:
  case SUBSET_NODE:
    for (i=Fp; i <= Lp; i++) {
      if ( Item[i][a].txt_index == 0 ) Swap(Fp++, i);
    }
    break;
  case CONTINUOUS_NODE:
    for (i=Fp; i <= Lp; i++) {
      if ( Item[i][a].real_value == Unknown ) 
	Swap(Fp++, i);
    }
    break;
  }
  
  return Fp - 1;
}


/*************************************************************************/
/*								 	 */
/*	Return the total weight of items from Fp to Lp		 	 */
/*								 	 */
/*************************************************************************/
ItemCount CountItems(int Fp,int Lp)
{
  register ItemCount Sum=0.0, *Wt, *LWt;

  if ( ! use_example_weights && AllKnown ) return Lp - Fp + 1;
  
  for ( Wt = Weight + Fp, LWt = Weight + Lp ; Wt <= LWt ; ) {
    Sum += *Wt++;
  }
  
  return Sum;
}



/*************************************************************************/
/*                                                               	 */
/*		Exchange items at a and b			 	 */
/*									 */
/*************************************************************************/
void Swap(int a,int b)
{
  register INSTANCE hold;
  register float    hold_weight;
  register float    tmp_example_weight; 

  hold = Item[a];
  Item[a] = Item[b];
  Item[b] = hold;
  
  hold_weight = Weight[a];
  Weight[a] = Weight[b];
  Weight[b] = hold_weight;

  if ( use_example_weights ) {
     tmp_example_weight = example_weight[a];
     example_weight[a]  = example_weight[b];
     example_weight[b]  = tmp_example_weight;
  }

  return;
}




