/*************************************************************************/
/*                                                                	 */
/*	Evaluation of a test on a continuous valued attribute	  	 */
/*	-----------------------------------------------------	  	 */
/*								  	 */
/*************************************************************************/
#include "buildex.h"


float *SplitGain, /* SplitGain[i] = gain with att value of item i as threshold */
      *SplitInfo;	/* SplitInfo[i] = potential info ditto */



/*************************************************************************/
/*								  	 */
/*  Continuous attributes are treated as if they have possible values	 */
/*	0 (unknown), 1 (less than cut), 2(greater than cut)	  	 */
/*  This routine finds the best cut for items Fp through Lp and sets	 */
/*  Info[], Gain[] and Bar[]						 */
/*								  	 */
/*************************************************************************/
void EvalContinuousAtt(int a, int Fp,int Lp)
{ 
  int i, BestI, Xp, Tries=0;
  float Items, KnownItems, LowItems, min_number_in_branch;
  int c;
  float AvGain=0, Val, BestVal, BaseInfo, ThreshCost;
  float lgini=0, rgini=0;
  
  Verbosity(2) printf("\tAtt %s", attribute_name[a]);
  Verbosity(3) printf("\n");
  
  reset_frequencies(2);
  
  /*  Omit and count unknown values */
  Items = CountItems(Fp, Lp);
  Xp = Fp;
  for (i=Fp; i <= Lp; i++) {
    if ( Item[i][a].real_value == Unknown ) {
      Freq[ 0 ][ Class(Item[i]) ] += Weight[i];
      Swap(Xp, i);
      Xp++;
    }
  }
  
  ValFreq[0] = 0;
  for (c=0; c < number_of_classes; c++) 
    ValFreq[0] += Freq[0][c];
  
  
  KnownItems = Items - ValFreq[0];
  UnknownRate[a] = 1.0 - KnownItems / Items;
  
  /*  Special case when very few known values  */
  if ( KnownItems < 2 * MINOBJS ) {
    Verbosity(2) printf("\tinsufficient cases with known values\n");
    Gain[a] = -Epsilon;
    Info[a] = 0.0;
    return;
  }
  
  Quicksort(Xp, Lp, a, Swap);
  
  /*  Count base values and determine base information  */
  for (i=Xp; i <= Lp; i++) {
    Freq[ 2 ][ Class(Item[i]) ] += Weight[i];
    SplitGain[i] = -Epsilon;
    SplitInfo[i] = 0;

  }
  
  if (use_gini) {
    for (c=0; c< number_of_classes;c++) 
      rgini+=Freq[2][c]*Freq[2][c];
  }


  BaseInfo = TotalInfo(Freq[2], 0, number_of_classes) / KnownItems;
  
  /*  Try possible cuts between items i and i+1, and determine the
      information and gain of the split in each case.  We have to be wary
      of splitting a small number of items off one end, as we can always
      split off a single item, but this has little predictive power.  */
  
  if (PURE_TREES)
     min_number_in_branch=MINOBJS;
  else
  {
    min_number_in_branch=0.10 * KnownItems / number_of_classes;
    if ( min_number_in_branch < MINOBJS ) 
      min_number_in_branch = MINOBJS;
    else if ( hard_limit_min_branch && min_number_in_branch > 25 ) 
      min_number_in_branch = 25;
  }
  LowItems = 0;
  for (i=Xp; i <= Lp - 1; i++) {
    c = Class(Item[i]);

    if ( use_gini ) {
      lgini+=Weight[i]*(2*Freq[1][c]+Weight[i]);
      rgini+=Weight[i]*(-2*Freq[2][c]+Weight[i]);
    }
    LowItems   += Weight[i];
    Freq[1][c] += Weight[i];
    Freq[2][c] -= Weight[i];


    /* Check if there are enough examples in the lower split. If not,
       skip to the next split. */
    if ( LowItems < min_number_in_branch ) 
      continue;
    /* Else if there aren't enough in the upper split, quit looking */
    else if ( LowItems > KnownItems - min_number_in_branch ) 
      break;

    if ( Item[i+1][a].real_value - Item[i][a].real_value > 1E-5 ) {
      ValFreq[1] = LowItems;
      ValFreq[2] = KnownItems - LowItems;

      if ( use_gini ) {
	SplitGain[i]=SplitInfo[i]=(lgini/ValFreq[1]) + (rgini/ValFreq[2]);
      } else {
	SplitGain[i] = ComputeGain(BaseInfo, UnknownRate[a], 2, KnownItems);
	SplitInfo[i] = TotalInfo(ValFreq, 0, 3) / Items;
      }
      AvGain += SplitGain[i];
      Tries++;
      
      Verbosity(3) {
	printf("\t\tCut at %.3f  (gain %.3f, val %.3f):",
	       ( CVal(Item[i],a) + CVal(Item[i+1],a) ) / 2,
	       SplitGain[i],
	       (double)Worth(SplitInfo[i], SplitGain[i], Epsilon));
	PrintDistribution(a, 2, True);
      }
    }

  }
  
  /*  Find the best attribute according to the given criterion  */
  ThreshCost = Log((double)Tries) / Items;
  if ( use_gini ) ThreshCost=0;
  BestVal = 0;
  BestI   = None;
  for (i=Xp; i <= Lp - 1; i++) {
    
    Val=SplitGain[i]-ThreshCost;
    if ( Val > BestVal ) {
      BestI   = i;
      BestVal = Val;
    }
    update_best_splits(a, Val, SplitInfo[i], Val,
		       (CVal(Item[i],a)+CVal(Item[i+1],a))/2.0, False);
  }

  /*  If a test on the attribute is able to make a gain,
      set the best break point, gain and information  */ 
  if ( BestI == None ) {
    Gain[a] = -Epsilon;
    Info[a] = 0.0;
    
    Verbosity(2) printf("\tno gain\n");
  } else {
    Bar[a]  = (Item[BestI][a].real_value + Item[BestI+1][a].real_value) / 2;
    Gain[a] = BestVal;
    Info[a] = SplitInfo[BestI];
    
    Verbosity(2)
      printf("\tcut=%.3f, inf %.3f, gain %.3f\n",
	     Bar[a], Info[a], Gain[a]);
  }
  
  return;
} 



/*************************************************************************/
/*                                                                	 */
/*  Change a leaf into a test on a continuous attribute           	 */
/*                                                                	 */
/*************************************************************************/
void ContinTest(TREE *node, int a)
{
  float Thresh;
  
  Sprout(node, 2);
  
  Thresh = GreatestValueBelow(a, Bar[a]);
  
  node->node_type=CONTINUOUS_NODE;
  node->tested_attribute=a;
  node->Cut		=
    node->Lower		=
    node->Upper		= Thresh;
  node->Errors        = 0;
  
  return;
}



/*************************************************************************/
/*                                                                	 */
/*  Return the greatest value of attribute Att below threshold t  	 */
/*                                                                	 */
/*************************************************************************/
float GreatestValueBelow(int a,float t)
{
  int i;
  float v, Best=0;
  Boolean NotYet=True;
  
  for (i=0; i <= MaxItem; i++) {
    v = CVal(Item[i], a);
    if ( v != Unknown && v <= t && ( NotYet || v > Best ) ) {
      Best = v;
      NotYet = False;
    }
  }
  
  return Best;
}
