/*************************************************************************/
/*									 */
/*	Calculate information, information gain, and print dists	 */
/*	--------------------------------------------------------	 */
/*									 */
/*************************************************************************/
#include "buildex.h"


/*************************************************************************/
/*									 */
/*  Determine the worth of a particular split according to the		 */
/*  operative criterion							 */
/*									 */
/*	    Parameters:							 */
/*		SplitInfo:	potential info of the split		 */
/*		SplitGain:	gain in info of the split		 */
/*		MinGain:	gain above which the Gain Ratio		 */
/*				may be used				 */
/*									 */
/*  If the Gain criterion is being used, the information gain of	 */
/*  the split is returned, but if the Gain Ratio criterion is		 */
/*  being used, the ratio of the information gain of the split to	 */
/*  its potential information is returned.				 */
/*									 */
/*************************************************************************/
float Worth(float ThisInfo, float ThisGain, float MinGain)
{
  if ( GAINRATIO ) {
    if ( ThisGain - MinGain >= -Epsilon && ThisInfo > Epsilon ) {
      return ThisGain / ThisInfo;
    } else {
      return -Epsilon;
    }
  } else if ( use_gini )
    return ThisGain;
  else
    return ( ThisInfo > 0 && ThisGain > -Epsilon ? ThisGain : -Epsilon );

}



/*************************************************************************/
/*									 */
/*  Given tables Freq[][] and ValFreq[], compute the information gain.	 */
/*									 */
/*	    Parameters:							 */
/*		BaseInfo:	average information for all items with	 */
/*				known values of the test attribute	 */
/*		UnknownRate:	fraction of items with unknown ditto	 */
/*		MaxVal:		number of forks				 */
/*		TotalItems:	number of items with known values of	 */
/*				test att				 */
/*									 */
/*  where Freq[x][y] contains the no. of cases with value x for a	 */
/*  particular attribute that are members of class y,			 */
/*  and ValFreq[x] contains the no. of cases with value x for a		 */
/*  particular attribute						 */
/*									 */
/*************************************************************************/
float ComputeGain(float BaseInfo, float UnknFrac, int MaxVal, float TotalItems)
{
  int v;
  float ThisInfo=0.0, ThisGain;
  short ReasonableSubsets=0;
  
  /*  Check whether all values are unknown or the same  */
  if ( ! TotalItems ) return -Epsilon;
  
  /*  There must be at least two subsets with MINOBJS items  */
  for (v=1; v <= MaxVal; v++) {
    if ( ValFreq[v] >= MINOBJS ) ReasonableSubsets++;
  }
  if ( ReasonableSubsets < 2 ) return -Epsilon;
  
  /*  Compute total info after split, by summing the
      info of each of the subsets formed by the test  */
  for (v=1; v <= MaxVal; v++) 
    ThisInfo += TotalInfo(Freq[v], 0, number_of_classes);
  
  /*  Set the gain in information for all items, adjusted for unknowns  */
  ThisGain = (1 - UnknFrac) * (BaseInfo - ThisInfo / TotalItems);
  
  Verbosity(5)
    printf("ComputeThisGain: items %.1f info %.3f base %.3f unkn %.3f result %.3f\n",
	   TotalItems + ValFreq[0], ThisInfo, BaseInfo, UnknFrac, ThisGain);
  
  return ThisGain;
}



/*************************************************************************/
/*									 */
/*  Compute the total information in V[ MinVal..MaxVal ]		 */
/*									 */
/*************************************************************************/
float TotalInfo(float V[], int min_value, int max_value)
{
  int v;
  float Sum=0.0;
  float N, TotalItems=0;

  for (v=min_value; v < max_value; v++) {
    N = V[v];
    
    Sum+= N * Log(N);
    TotalItems += N;
  }

  return TotalItems * Log(TotalItems) - Sum;

}


/*************************************************************************/
/*									 */
/*	Print distribution table for given attribute			 */
/*									 */
/*************************************************************************/
void PrintDistribution(int a, int MaxVal, Boolean ShowNames)
{
  int v, c;
  char *Val;
  
  printf("\n\t\t\t ");
  for (c=0; c <= MaxClass; c++) 
    printf("%7.6s", class_name[c]);
  printf("\n");
  
  for (v=0; v <= MaxVal; v++) {
    if ( ShowNames ) {
      Val = ( !v ? "unknown" :
	      attribute_type[a]==DISCRETE ? AttValName[a][v] :
	      v == 1 ? "below" : "above" );
      printf("\t\t[%-7.7s:", Val);
    } else {
      printf("\t\t[%-7d:", v);
    }
    
    for (c=0; c <= MaxClass; c++) 
      printf(" %6.1f", Freq[v][c]);
    printf("]\n");
  }
  return;
}
