/*
  predict.c

  Output predictions for the examples. This will work with multiple
  trees (by majority vote) and for the simple case of one tree.

  The class label is output to a .pred file, one line per file. The
  weighted prediction for each class (represented as the sum of examples
  of that class at the leaf node) is also output to a .weighted_pred
  file.
*/

#include "defns.h"
#include "types.h"
#include "extern.h"

/* Additional external variables needed here (in addition to extern.h) */
extern float    *ClassSum;
extern Boolean  PRINT_ALL_TREES;

void predict(Boolean resubstitution)
{
  int pred,i,c,t,leaf;
  FILE *fp,*wfp, *lfp, *lafp;
  char buf[1024];
  Tree *trees;
  int *vote;
  float *weighted_vote, *BoostingVote;
  int correct_predictions=0;

  if ( resubstitution )
    load_training_data();
  else 
    load_test_data();

  /* Set up some variables, etc. */
  trees=(use_pruning)?pruned_trees:unpruned_trees;
  vote=calloc(number_of_classes,sizeof(int));
  weighted_vote=calloc(number_of_classes,sizeof(float));

  if(PRINT_ALL_TREES)
  {
    for(t=0;t<number_of_trees;t++)
    {
      sprintf(buf,"%s-%d.pred",filestem,(t+1));
      if ( (fp=fopen(buf,"w")) == 0 )
      {
        fprintf(stderr,"Warning: Open of %s failed, no predictions made.",buf);
        return;
      }
      for(i=0;i<number_of_examples;i++)
        fprintf(fp,"%s\n",class_name[Category(Item[i], trees[t])]);
      fclose(fp);
    }
  }

  /* Open the files */
  sprintf(buf,"%s.pred",filestem);
  if ( (fp=fopen(buf,"w")) == 0 ) {
    fprintf(stderr,"Warning: Open of %s failed, no predictions made.",buf);
    return;
  }
  sprintf(buf,"%s.laplace", filestem);
  if ( (lafp=fopen(buf,"w")) == 0 ) {
    fprintf(stderr,"Warning: Open of %s failed, no predictions made.",buf);
    return;
  }
  sprintf(buf,"%s.weighted_pred", filestem);
  if ( (wfp=fopen(buf,"w")) == 0 ) {
    fprintf(stderr,"Warning: Open of %s failed, no predictions made.",buf);
    return;
  }
  /* In weighted prediction file, first line consists of class labels */
  for (c=0; c < number_of_classes; c++) {
    fprintf(wfp,"%s",class_name[c]);
    fprintf(wfp,"%c",(c==number_of_classes-1)?'\n':',');
  }
  if(OUTPUT_LEAFS)
    {
      sprintf(buf,"%s.leafs",filestem);
      if ( (lfp=fopen(buf,"w")) == 0 ) {
	fprintf(stderr,"Warning: Open of %s failed, no predictions made.",buf);
	return;
      }
      fclose(lfp);
    }
  if ( use_boosting )
  {
    BoostingVote = malloc(number_of_classes * sizeof(float));
    for(i=0;i<number_of_examples;i++)
    {
      for(c=0;c<number_of_classes;c++)
        BoostingVote[c]=0;
      for(t=0;t<number_of_trees;t++)
      {
        pred=Category(Item[i], trees[t]);
        BoostingVote[pred]+=BoostingClassifierWeights[t];
      }
      for (pred=0,c=0; c < number_of_classes; c++)
        if ( BoostingVote[c] > BoostingVote[pred] ) pred=c;
      fprintf(fp,"%s\n",class_name[pred]);
      if ( Class(Item[i]) == pred ) correct_predictions++;
    }
  }
  else
  {
    /* For each example, test it on the tree(s) and record the vote */
    for (i=0; i < number_of_examples; i++) {
      /* Zero out votes */
      for (c=0; c < number_of_classes; c++) 
        weighted_vote[c]=vote[c]=0;

      /* Accumulate votes from each tree */
      for (t=0; t < number_of_trees; t++) {
        pred=Category(Item[i], trees[t]);
        vote[pred]++;
        for (c=0; c < number_of_classes;c++)
    	  weighted_vote[c]+=ClassSum[c];
      }

 
      /* Get the majority vote, then print out to a file */
      for (pred=0,c=0; c < number_of_classes; c++)
        if ( vote[c] > vote[pred] ) pred=c;
      fprintf(fp,"%s\n",class_name[pred]);

     
      //if(OUTPUT_LEAFS)
      //{
	  /* Get the Tree Leaf ID, then print out to a file  */
	  //leaf=GetLeafID(Item[i], trees[0]);
	  //fprintf(lfp, "%d\n",leaf);
      //}
      /* Next, calculate the weighted votes (average of individual trees) */
      for (c=0; c < number_of_classes; c++) {
        fprintf(wfp,"%.1f", weighted_vote[c]/number_of_trees);
        fprintf(wfp,"%c",(c==number_of_classes-1)?'\n':',');
      }

      /* And accumulate the accuracy */
      if ( Class(Item[i]) == pred ) correct_predictions++;
    
    }
  }
  /* Clean up */
  fclose(fp);
  fclose(wfp);
  free(vote);
  free(weighted_vote);


  /* Print out the accuracy summary */
  printf("\nAccuracy: %5.2f%%\n",100.0 * 
	           (correct_predictions/(float)number_of_examples));

  return;
}
