/*************************************************************************/
/*									 */
/*	Prune a decision tree and predict its error rate		 */
/*	------------------------------------------------		 */
/*									 */
/*************************************************************************/
#include "defns.h"
#include "types.h"
#include "extern.h"

float EstimateErrors(TREE *T, int Fp,int Lp, int Sh,Boolean UpdateTree);
void CheckPossibleValues();

extern	float	*Weight;

Set	*PossibleValues=NULL;
Boolean	Changed;

#define	LocalVerbosity(x)	if (Sh >= 0 && VERBOSITY >= x)
void Indent();
#define	Intab(x)		Indent(x, "| ")



TREE *prune_tree(TREE *tree)
{
  TREE *p;

  if ( VERBOSITY > 0 ) info("Pruning tree\n");

  /* We can prune on the validation set */
  if ( use_validation_set )
    load_validation_data();

  p = CopyTree(tree);
  if ( Prune(p) && number_of_trees == 1 ) {
    printf("\nSimplified ");
    print_tree(p);
  }
  
  
  return p;
}


/*************************************************************************/
/*									 */
/*  Prune tree T, returning true if tree has been modified		 */
/*									 */
/*************************************************************************/
Boolean Prune(TREE *T)
{
  int i,a;

  InitialiseWeights();
  AllKnown = True;
  
  Verbosity(1) printf("\n");
  
  Changed = False;
  
  EstimateErrors(T, 0, number_of_examples, 0, True);
  
  if ( SUBSET ) {
    if ( ! PossibleValues ) 
      PossibleValues = (Set *) calloc(number_of_attributes, sizeof(Set));

    for (a=0; a < number_of_attributes; a++) {
      if ( attribute_type[a] == DISCRETE || attribute_type[a]==BINARY ) {
	PossibleValues[a] = (Set) malloc((MaxAttVal[a]>>3) + 1);
	ClearBits((MaxAttVal[a]>>3) + 1, PossibleValues[a]);

	for (i=1; i <= MaxAttVal[a]; i++) {
	  SetBit(i, PossibleValues[a]);
	}
      }
    }

    CheckPossibleValues(T);
  }
  
  return Changed;
}




/*************************************************************************/
/*									 */
/*	Estimate the errors in a given subtree				 */
/*									 */
/*************************************************************************/
float EstimateErrors(TREE *T, int Fp,int Lp, int Sh,Boolean UpdateTree)
{ 
  int i, Kp, Ep;
  float Cases, KnownCases,  TreeErrors, LeafErrors,
    ExtraLeafErrors, BranchErrors, Factor, MaxFactor;
  float *class_distribution;
  int v, MaxBr=1;
  int c, best_class;
  Boolean PrevAllKnown;

  Lp--;

  

  Cases = CountItems(Fp, Lp);

  /*  Generate the class frequency distribution  */
  class_distribution=CALLOC(number_of_classes, sizeof(float));
  for (i=Fp; i <= Lp; i++) 
    class_distribution[ Class(Item[i]) ] += Weight[i];

  /*  Find the most frequent class and update the tree  */
  best_class = T->leaf;
  for (c=0; c < number_of_classes; c++) {
    if ( class_distribution[c] > class_distribution[best_class] ) 
      best_class = c;
  }
  LeafErrors = Cases - class_distribution[best_class];
  ExtraLeafErrors = AddErrs(Cases, LeafErrors);
  


  if ( UpdateTree ) {
    T->Items = Cases;
    T->leaf  = best_class;
    for (i=0; i < number_of_classes; i++) 
      T->class_distribution[i]=class_distribution[i];
  }

  if ( T->node_type == LEAF_NODE ) {	/*  leaf  */
    TreeErrors = LeafErrors + ExtraLeafErrors;
    
    if ( UpdateTree ) {
      T->Errors = TreeErrors;
      
      if ( VERBOSITY >= 1 && Sh >= 0 ) {
	  Intab(Sh);
	  info("%s (%.2f:%.2f/%.2f)\n", class_name[T->leaf],
		 T->Items, LeafErrors, T->Errors);
      }
    }

    free(class_distribution);
    return TreeErrors;
  }

  /*  Estimate errors for each branch  */
  Kp = group_unknown_values(Fp, Lp, T) + 1;
  KnownCases = CountItems(Kp, Lp);

  PrevAllKnown = AllKnown;
  if ( Kp != Fp ) AllKnown = False;
  
  TreeErrors = MaxFactor = 0;
  
  for (v=1; v <= T->num_branches; v++) {
    Ep = Group(v, Kp, Lp, T);
    
    /* If there are examples for this branch */
    if ( Kp <= Ep ) {
      Factor = CountItems(Kp, Ep) / KnownCases;
      
      if ( Factor >= MaxFactor ) {
	MaxBr = v;
	MaxFactor = Factor;
      }
      
      for (i=Fp; i <= Kp-1; i++) 
	Weight[i] *= Factor;

      TreeErrors += EstimateErrors(T->branch[v], Fp, Ep+1, Sh+1, UpdateTree);

      group_unknown_values(Fp, Ep, T);
      for (i=Fp; i <= Kp-1; i++) 
	Weight[i] /= Factor;
    } else {
      /* Else there are no examples. This should only happen in the case
	 of using a validation set (check that here). */
      if ( use_validation_set && T->branch[v]->node_type != LEAF_NODE ) {
	/* If this is a sub-tree (not a leaf) then prune back to a leaf */
	T->branch[v]->node_type=LEAF_NODE;
	T->branch[v]->leaf=best_class;
	T->branch[v]->Items=0;
	T->branch[v]->Errors=0;
      }
    }
  }    
  AllKnown = PrevAllKnown;
  
  if ( ! UpdateTree ) {
    free(class_distribution);
    return TreeErrors;
  }

  /*  See how the largest branch would fare  */
  BranchErrors = EstimateErrors(T->branch[MaxBr], Fp, Lp+1, -1000, False);

  LocalVerbosity(1) {
    Intab(Sh);
    printf("%s:  [%d%%  N=%.2f  tree=%.2f  leaf=%.2f+%.2f  br[%d]=%.2f]\n",
	   attribute_name[T->tested_attribute],
	   (int) ((TreeErrors * 100) / (T->Items + 0.001)),
	   T->Items, TreeErrors, LeafErrors, ExtraLeafErrors,
	   MaxBr, BranchErrors);
  }

  /*  See whether tree should be replaced with leaf or largest branch  */
  if ( LeafErrors + ExtraLeafErrors <= BranchErrors + 0.1 &&
       LeafErrors + ExtraLeafErrors <= TreeErrors + 0.1 ) {
    LocalVerbosity(1) {
      Intab(Sh);
      printf("Replaced with leaf %s\n", class_name[T->leaf]);
    }
    
    T->node_type = 0;
    T->Errors = LeafErrors + ExtraLeafErrors;
    Changed = True;
  } else if ( BranchErrors <= TreeErrors + 0.1  && use_subtree_raising ) {
    LocalVerbosity(1) {
      Intab(Sh);
      printf("Replaced with branch %d\n", MaxBr);
    }
    
    AllKnown = PrevAllKnown;
    EstimateErrors(T->branch[MaxBr], Fp, Lp+1, Sh, True);
    memcpy((void *)T, (void *)T->branch[MaxBr], sizeof(TREE));
    Changed = True;
  } else {
    T->Errors = TreeErrors;
  }
  
  AllKnown = PrevAllKnown;
  free(class_distribution);
  
  return T->Errors;
}



/*************************************************************************/
/*									 */
/*	Remove unnecessary subset tests on missing values		 */
/*									 */
/*************************************************************************/
void CheckPossibleValues(Tree T)
{
  Set HoldValues;
  int v, Bytes, b;
  int a;
  char Any=0;
  
  if ( T->node_type == BrSubset ) {
    a = T->tested_attribute;
    
    Bytes = (MaxAttVal[a]>>3) + 1;
    HoldValues = (Set) malloc(Bytes);

    /*  See if last (default) branch can be simplified or omitted  */
    for (b=0; b <= Bytes-1; b++) {
      T->Subset[T->num_branches][b] &= PossibleValues[a][b];
      Any |= T->Subset[T->num_branches][b];
    }

    if ( ! Any ) 
      T->num_branches--;

    /*  Process each subtree, leaving only values in branch subset  */
    CopyBits(Bytes, PossibleValues[a], HoldValues);
    for (v=1; v<=T->num_branches; v++) {
      CopyBits(Bytes, T->Subset[v], PossibleValues[a]);
      
      CheckPossibleValues(T->branch[v]);
    }
    
    CopyBits(Bytes, HoldValues, PossibleValues[a]);
    
    free(HoldValues);
  } else if ( T->node_type ) {
    for (v=1; v <= T->num_branches; v++) 
      CheckPossibleValues(T->branch[v]);
  }

  return;
}
