#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <ctype.h>
#include <zlib.h>
#include "utils.h"
#include "types.h"
#include "extern.h"


void _output_tree();
void _output_discrete_names();
void _save_trees(TREE **T, int tree_type);

TREE **_load_trees();
TREE *_input_tree();
void _input_discrete_names();



void save_pruned_trees(TREE **T)
{
  if ( VERBOSITY > 0 ) info("Saving pruned trees\n");
  _save_trees(T,PRUNED_TREE);
}

void save_unpruned_trees(TREE **T)
{
  if ( VERBOSITY > 0 ) info("Saving unpruned trees\n");
  _save_trees(T,UNPRUNED_TREE);
}

TREE **load_pruned_trees()
{
  if ( VERBOSITY > 0 ) info("Loading pruned trees\n");
  return _load_trees(PRUNED_TREE);
}

TREE **load_unpruned_trees()
{
  if ( VERBOSITY > 0 ) info("Loading unpruned trees\n");
  return _load_trees(UNPRUNED_TREE);
}






void _save_trees(TREE **T, int tree_type)
{
  char filename[MAXBUF];
  gzFile tfp;
  long tt;
  int i;

  /* Create appropriate filename */
  if ( tree_type == PRUNED_TREE )
    sprintf(filename,"%s.pruned_tree.gz", filestem);
  else
    sprintf(filename,"%s.unpruned_tree.gz", filestem);

  tfp=GZOPEN(filename,"w");
  time(&tt);
  gzprintf(tfp,"# \n# Created on %s# filestem: %s\n# %s tree\n",
	  ctime(&tt), filestem, 
	  (tree_type == PRUNED_TREE)?"Pruned":"Unpruned");


  gzprintf(tfp,"\nnumber_of_trees=%d\n\n", number_of_trees);
  for (i=0; i < number_of_trees; i++) {
    
    gzprintf(tfp,"\n#BEGIN_TREE\nTreeID=%d\n\n", i);
    _output_tree(tfp,T[i]);
    gzprintf(tfp,"#END_TREE\n");
  }

  _output_discrete_names(tfp);

  gzclose(tfp);
  
  return;

}


void _output_tree(gzFile tfp, TREE *T)
{
  int i,v,Bytes;

  gzprintf(tfp, "#BEGIN_NODE\n");
  gzprintf(tfp, "NodeID=%d\n", T->nodeid);
  gzprintf(tfp, "node_type=%s\n",node_type_string[T->node_type]);
  gzprintf(tfp, "Class=%s\n", class_name[T->leaf]);
  gzprintf(tfp, "Items=%f\n", T->Items);
  gzprintf(tfp, "Errors=%f\n",T->Errors);
  for (i=0; i < number_of_classes; i++) 
    gzprintf(tfp,"Frequency[%s]=%f\n", class_name[i], T->class_distribution[i]);

  if ( T->node_type != LEAF_NODE ) {
    gzprintf(tfp,"SplitAttribute=%s\n", attribute_name[T->tested_attribute]);
    gzprintf(tfp,"Branches=%d\n", T->num_branches);

    /* Dump specific info about the branches */
    for (v=1; v <= T->num_branches; v++) {
      gzprintf(tfp,"BranchID[%d]=%d\n",v,T->branch[v]->nodeid);
      if ( T->node_type == CONTINUOUS_NODE ) {
	gzprintf(tfp,"Branch[%d]=%s\n",v,(v==1)?"<=":">");
      } else if ( T->node_type == DISCRETE_NODE) {
	gzprintf(tfp,"Branch[%d]=%s\n",v,AttValName[T->tested_attribute][v]);
      }
    }

    switch (T->node_type) {
    case DISCRETE_NODE:
      break;
    case CONTINUOUS_NODE:
      gzprintf(tfp,"Cut=%f\n",T->Cut);
      gzprintf(tfp,"Lower=%f\n",T->Lower);
      gzprintf(tfp,"Upper=%f\n", T->Upper);
      break;
    case SUBSET_NODE:
      Bytes=(MaxAttVal[T->tested_attribute]>>3) + 1;
      for (v=1; v <= T->num_branches; v++) {
	gzprintf(tfp,"Subset[%d]=",v);
	for (i=0; i < Bytes; i++) 
	  gzprintf(tfp,"%x",T->Subset[v][i]);
	gzprintf(tfp,"\n");
      }
      break;
    }

  }
  
  gzprintf(tfp, "#END_NODE\n\n");
  
  if ( T->node_type != LEAF_NODE ) {
     for (v=1; v <= T->num_branches; v++)
       _output_tree(tfp,T->branch[v]);
  }

  return;
}



void _output_discrete_names(gzFile tfp)
{
  int a,v;


  for (a=0; a < number_of_attributes; a++) {
    if ( attribute_type[a] != DISCRETE )
      continue;
    gzprintf(tfp, "AttributeCount[%d]=%d\nAttributeValues[%d]=", 
	    a, MaxAttVal[a],a);
    for (v=1; v <= MaxAttVal[a]; v++) {
      gzprintf(tfp,"%s%c", AttValName[a][v], v==MaxAttVal[a]?'\n':',');
    }

  }

  return;
}




/*=====================================================================
  load_tree()

  Recover tree(s) from  appropriately named text files.

  =====================================================================*/

TREE **_load_trees(int tree_type)
{
  TREE **t;
  char filename[MAXBUF];
  gzFile tfp;
  int i;
  char tbuf[MAXBUF];
  int tmpint;

  /* Create appropriate filename */
  if ( tree_type == PRUNED_TREE )
    sprintf(filename,"%s.pruned_tree", filestem);
  else
    sprintf(filename,"%s.unpruned_tree", filestem);
  tfp=GZOPEN(filename,"r");

  /* Get the tree and discrete names, in correct order */
  while ( gzgets(tfp,tbuf,MAXBUF) != NULL ) 
    if ( sscanf(tbuf,"number_of_trees=%d", &tmpint) == 1 )  {
	number_of_trees=tmpint;
        break;
    }
  if (gzeof(tfp)) die("Could not recover the number of trees in file\n");
  
  /* Go through text file and get the trees */
  t=CALLOC(number_of_trees,sizeof(TREE *));

  for (i=0; i < number_of_trees; i++) {
    while ( gzgets(tfp,tbuf,MAXBUF) != NULL ) 
      if ( sscanf(tbuf,"TreeID=%d", &tmpint) == 1 ) 
	break;
      
    if (gzeof(tfp) || tmpint != i ) 
      die("Problem with tree id in file\n");

    t[i]=_input_tree(tfp);
    if ( VERBOSITY > 0 ) 
       print_tree(t[i]);
  }

  
  /* Discrete names are listed at the end of the file */
  _input_discrete_names(tfp);

  gzclose(tfp);

  return t;
}

/* Generic routine to recursively load tree from file */
TREE *_input_tree(gzFile tfp)
{

  TREE *t;
  float tmpfloat;
  int tmpint;
  char tmpbuf[MAXBUF], tbuf[MAXBUF];
  int i;


  t=malloc(sizeof(TREE));
  t->class_distribution=CALLOC(number_of_classes,sizeof(float));
  t->node_type=-1;

  while (gzgets(tfp,tbuf,MAXBUF) != NULL) {
    tbuf[strlen(tbuf)-1]=0;
    /* One case in which we break out of this loop */
    if (!strncmp("#END_NODE", tbuf,9) ) break;

    if (sscanf(tbuf,"NodeID=%d", &tmpint) == 1 ) {
       t->nodeid=tmpint;
       continue;
    }
 
    /* Get node type */
    if ( sscanf(tbuf,"node_type=%s", tmpbuf) == 1 ) {
      for (i=0; i < 4; i++) 
	if ( !strcasecmp(tmpbuf,node_type_string[i])) 
	  break;
      if ( i == 4 ) die("Could not recover node type\n");
      t->node_type=i;
      continue;
     }

     /* Get the default class */
     if ( sscanf(tbuf,"Class=%s", tmpbuf) == 1 )  {
        for (i=0; i < number_of_classes;i++) 
           if (!strcasecmp(tmpbuf, class_name[i])) 
              break;
        if ( i == number_of_classes ) 
           die("Could not recover class for node\n");
        t->leaf=i;
        continue;
     }

     if ( sscanf(tbuf,"Items=%f", &(t->Items)) == 1 ) continue;
     if ( sscanf(tbuf,"Errors=%f",&(t->Errors)) == 1) continue;


     /* Split attribute */
     if ( t->node_type != LEAF_NODE && 
	  !strncmp(tbuf,"SplitAttribute=",15)) {
       char *p=tbuf+15;
       for (i=0; i < number_of_attributes;i++) 
	 if (!strcasecmp(p,attribute_name[i])) break;
       if (i==number_of_attributes) die("Trouble getting attribute name\n");
       t->tested_attribute=i;
       continue;
     }


     /* Number of branches */
     if ( t->node_type != LEAF_NODE && 
	  sscanf(tbuf,"Branches=%d", &tmpint) == 1 ) {
       t->num_branches=tmpint;
       continue;
     }

     /* Continuous node information */
     if ( t->node_type==CONTINUOUS_NODE && 
	  sscanf(tbuf,"Cut=%f", &(t->Cut))==1)
       continue;
     if ( t->node_type==CONTINUOUS_NODE && 
	  sscanf(tbuf,"Lower=%f", &(t->Lower))==1)
       continue;
     if ( t->node_type==CONTINUOUS_NODE && 
	  sscanf(tbuf,"Upper=%f", &(t->Upper))==1)
       continue;
     if ( t->node_type==SUBSET_NODE )
       die("Subset attributes not supported yet\n");

     /* Class Frequencies */
     if ( !strncmp("Frequency",tbuf,9) ) {
       char *p,*q;
       int c;

       /* Nasty thing to get [p]=q correct. */
       p=q=tbuf+10;
       while (*q != ']') q++;
       tmpfloat=atof(q+2);
       *q=0;
       for (c=0; c < number_of_classes; c++) 
	 if (!strcmp(p,class_name[c])) 
	   break;
       if (c == number_of_classes ) die("Could not find the class\n");
       t->class_distribution[c]=tmpfloat;
       continue;
     }


     /* This is the default case. For debugging, you can print out
	extra junk, or leave empty. */
     //info("Default: %s was not used.\n",tbuf);

  }

  /* After the node has been read in, we can recursively call ourselves
     to add subtrees */
  if ( t->node_type != LEAF_NODE ) {
    int v;

    t->branch=CALLOC(t->num_branches+1,sizeof(TREE *));
    for (v=1;v <= t->num_branches; v++)
      t->branch[v]=_input_tree(tfp);
  }

  return t;
}


void _input_discrete_names(gzFile tfp)
{
  int c;
  char tmpbuf[MAXBUF],tbuf[MAXBUF];

  while (gzgets(tfp,tbuf,MAXBUF) != NULL ) {
    if ( !strncmp(tbuf,"AttributeCount", 14) ) {
       char *p,*q;
       int attr, count;

       p=q=tbuf+15;
       while ( *q != ']' ) q++;
       count=atoi(q+2);
       *q=0;
       attr=atoi(p);

       if ( attribute_type[attr] != DISCRETE )
          die("Got attribute values for non-discrete attribute!\n");

       /* Get the number of attribute values for this attribute */
       MaxAttVal[attr]=count;

       /* Now get each value, in order */
       while(gzgetc(tfp) != '=');
       c=1;
       while ( c <= MaxAttVal[attr] ) {
          p=tmpbuf;
          while ((*p=gzgetc(tfp)) != ',' && *p != '\n') p++;
          *p=0;
          AttValName[attr][c++]=strdup(tmpbuf);
       }
    }
  }

  return;
}

