/*************************************************************************/
/*									 */
/*	Routines for displaying, building, saving and restoring trees	 */
/*	-------------------------------------------------------------	 */
/*									 */
/*************************************************************************/
#include "defns.h"
#include "types.h"
#include "extern.h"

void RecoverDiscreteNames();
void  SaveDiscreteNames();
void StreamIn();
void StreamOut();
void OutTree();
void Indent();

#define	Tab		"|   "
#define	TabSize		4
#define	Width		80	/* approx max width of printed trees */

/*  If lines look like getting too long while a tree is being
    printed, subtrees are broken off and printed separately after
    the main tree is finished	 */

int	Subtree;		/* highest subtree to be printed */
Tree	Subdef[100];		/* pointers to subtrees */

FILE	*TRf = 0;	/* file pointer for tree i/o */
char	filename[500];		/* file name */


void Show(TREE *T,int Sh);
void ShowBranch(int Sh,TREE *T,int v);
void Indent(int Sh,char *Mark);
int MaxLine(TREE *St);

/*************************************************************************/
/*									 */
/*	Display entire decision tree T					 */
/*									 */
/*************************************************************************/
void print_tree(TREE *T)
{
  int s;
  
  Subtree=0;
  printf("Decision Tree:\n");
  Show(T, 0);
  printf("\n");
  
  for (s=1; s <= Subtree; s++) {
    printf("\n\nSubtree [S%d]\n", s);
    Show(Subdef[s], 0);
    printf("\n");
  }
  printf("\n");
  return;
}



/*************************************************************************/
/*									 */
/*	Display the tree T with offset Sh				 */
/*									 */
/*************************************************************************/
void Show(TREE *t,int Sh)
{
  int v, MaxV;
  
  if ( t->node_type != LEAF_NODE ) {
    /*  See whether separate subtree needed  */
    if ( t != NULL && Sh && Sh * TabSize + MaxLine(t) > Width ) {
      if ( Subtree < 99 ) {
	Subdef[++Subtree] = t;
	printf("[S%d]", Subtree);
      } else {
	printf("[S??]");
      }
    } else {
      MaxV = t->num_branches;
      
      /*  Print simple cases first */
      for (v=1; v <= t->num_branches; v++) {
	if ( t->branch[v]->node_type == LEAF_NODE ) 
	  ShowBranch(Sh, t, v);
      }

      /*  Print subtrees  */
      for (v=1; v <= MaxV; v++) {
	if ( t->branch[v]->node_type != LEAF_NODE ) 
	  ShowBranch(Sh, t, v);
      }
    }
  } else {
    printf(" %s (%.1f", class_name[t->leaf], t->Items);
    if ( t->Errors > 0 ) printf("/%.1f", t->Errors);
    printf(")");
  }
  return;
}



/*************************************************************************/
/*									 */
/*	Print a node T with offset Sh, branch value v, and continue	 */
/*									 */
/*************************************************************************/
void ShowBranch(int Sh,TREE *T,int v)
{
  int Pv, Last=0;
  int a;
  Boolean FirstValue;
  int TextWidth, Skip, Values=0, i;
  
  a = T->tested_attribute;
  
  switch ( T->node_type ) {
  case DISCRETE_NODE:
    Indent(Sh, Tab);
    printf("%s = %s:", attribute_name[a], AttValName[a][v]);
    break;

  case CONTINUOUS_NODE:
    Indent(Sh, Tab);
    printf("%s %s %g ",
	   attribute_name[a], ( v == 1 ? "<=" : ">" ), T->Cut);
    if ( T->Lower != T->Upper ) 
      printf("[%g,%g]", T->Lower, T->Upper);
    printf(":");
    break;

  case SUBSET_NODE:
    /*  Count values at this branch  */
    for (Pv=1; Pv <= MaxAttVal[a]; Pv++) {
      if ( In(Pv, T->Subset[v]) ) {
	Last = Pv;
	Values++;
      }
    }
    if ( ! Values ) return;
    Indent(Sh, Tab);
    if ( Values == 1 ) {
      printf("%s = %s:", attribute_name[a], AttValName[a][Last]);
      break;
    }
    printf("%s in {", attribute_name[a]);
    FirstValue = True;
    Skip = TextWidth = strlen(attribute_name[a]) + 5;
    
    for (Pv=1; Pv <= MaxAttVal[a]; Pv++) {
      if ( In(Pv, T->Subset[v]) ) {
	if ( ! FirstValue &&
	     TextWidth + strlen(AttValName[a][Pv]) + 11 > Width ) {
	  Indent(Sh, Tab);
	  for (i=1; i <= Skip; i++) putchar(' ');
	  TextWidth = Skip;
	  FirstValue = True;
	}

	printf("%s%c", AttValName[a][Pv], Pv == Last ? '}' : ',');
	TextWidth += strlen(AttValName[a][Pv]) + 1;
	FirstValue = False;
      }
    }
    putchar(':');
  }
  
  Show(T->branch[v], Sh+1);

  return;
}



/*************************************************************************/
/*									 */
/*	Find the maximum single line size for non-leaf subtree St.	 */
/*	The line format is						 */
/*			<attribute> <> X.xx:[ <class (<Items>)], or	 */
/*			<attribute> = <DVal>:[ <class> (<Items>)]	 */
/*									 */
/*************************************************************************/
int MaxLine(TREE *St)
{
  int a;
  int v,MaxV, Next, Ll, MaxLl=0;

  
  a = St->tested_attribute;
  
  MaxV = St->num_branches;
  for (v=1; v <= MaxV; v++) {
    Ll = ( St->node_type == 2 ? 4 : strlen(AttValName[a][v]) ) + 1;
    
    /*  Find the appropriate branch  */
    Next = v;

    if ( ! St->branch[Next]->node_type ) {
      Ll += strlen(class_name[St->branch[Next]->leaf]) + 6;
    }
    MaxLl = Max(MaxLl, Ll);
  }

  return strlen(attribute_name[a]) + 4 + MaxLl;
}



/*************************************************************************/
/*								   	 */
/*	Indent s columns					  	 */
/*								  	 */
/*************************************************************************/
void Indent(int s,char *mark)
{
  printf("\n");
  while ( s-- ) printf("%s", mark);
}



/*************************************************************************/
/*									 */
/*	Free up space taken up by tree Node				 */
/*									 */
/*************************************************************************/
void ReleaseTree(TREE *node)
{
  int v;

  if ( node->node_type != LEAF_NODE ) {
    for (v=1; v <= node->num_branches; v++) 
      ReleaseTree(node->branch[v]);
    
    
    free(node->branch);
    
    if ( node->node_type == SUBSET_NODE ) 
      free(node->Subset);
  }
  
  free(node->class_distribution);
  free(node);
  
  return;
}



/*************************************************************************/
/*									 */
/*	Construct a leaf in a given node				 */
/*									 */
/*************************************************************************/
TREE *create_leaf(int nodeid,float *ClassFreq, 
			int node_class, 
			float num_cases, 
			float errors)
{
  TREE *node;
  int i;

  node=CALLOC(1, sizeof(TREE));
  node->class_distribution=CALLOC(number_of_classes, sizeof(float));
  for (i=0; i < number_of_classes; i++) 
    node->class_distribution[i]=ClassFreq[i];

  node->nodeid=nodeid;
  node->node_type = LEAF_NODE; 
  node->leaf = node_class;
  node->Items = num_cases;
  node->Errors = errors;
  node->num_branches=0; 
  return node; 
}



/*************************************************************************/
/*									 */
/*	Insert branches in a node 	                 		 */
/*									 */
/*************************************************************************/
void Sprout(TREE *node, int number_of_branches)
{
  node->num_branches = number_of_branches;
    
  node->branch=CALLOC(number_of_branches+1, sizeof(TREE *));
   
  return;
}



/*************************************************************************/
/*									 */
/*	Count the nodes in a tree					 */
/*									 */
/*************************************************************************/
int TreeSize(TREE *node)
{
  int sum=0;
  int  v;

  if ( node->node_type != LEAF_NODE ) {
    for (v=1; v <= node->num_branches; v++) {
      sum += TreeSize(node->branch[v]);
    }
  }

  return sum + 1;
}



/*************************************************************************/
/*									 */
/*	Return a copy of tree T						 */
/*									 */
/*************************************************************************/
TREE *CopyTree(TREE *T)
{
  int v,i;
  TREE *new_tree;
  

  new_tree=MALLOC(sizeof(TREE));
  memcpy(new_tree, T, sizeof(TREE));

  new_tree->class_distribution=CALLOC(number_of_classes, sizeof(float));
  for (i=0; i < number_of_classes; i++) 
    new_tree->class_distribution[i]=T->class_distribution[i];
  
  if ( T->node_type != LEAF_NODE ) {
    new_tree->branch=CALLOC(T->num_branches + 1, sizeof(TREE *));
    for (v=1; v <= T->num_branches; v++)
      new_tree->branch[v]=CopyTree(T->branch[v]);
  }
  
  return new_tree;
}


