/*
  windowing.c

  Code for handling windowing in C4.5. It is no longer the default
  approach and is so relegated to a separate file to avoid clutter.

*/

#include "types.h"
#include "defns.h"
#include "extern.h"


/* Global variable for file */
static int  *TargetClassFreq;

/* Functions */
int  BestTree();
Tree Iterate();
void FormInitialWindow();
void FormTarget();

/*************************************************************************/
/*									 */
/*	Grow and prune TRIALS trees and select the best of them		 */
/*									 */
/*************************************************************************/
int BestTree()
{
  int t, Best=0;
  
  InitialiseTreeData();
  
  TargetClassFreq = (int *) calloc(MaxClass+1, sizeof(int));
  
  unpruned_trees   = (Tree *) calloc(TRIALS, sizeof(Tree));
  pruned_trees = (Tree *) calloc(TRIALS, sizeof(Tree));
  
  /*  If necessary, set initial size of window to 20% (or twice
      the sqrt, if this is larger) of the number of data items,
      and the maximum number of items that can be added to the
      window at each iteration to 20% of the initial window size  */
  
  if ( ! WINDOW ) 
    WINDOW = Max(2 * sqrt(MaxItem+1.0), (MaxItem+1) / 5);
  
  
  if ( ! INCREMENT ) 
    INCREMENT = Max(WINDOW / 5, 1);
  
  
  FormTarget(WINDOW);
  
  /*  Form set of trees by iteration and prune  */
  for (t=0; t < TRIALS; t++) {
    FormInitialWindow();
    
    printf("\n--------\nTrial %d\n--------\n\n", t);
    
    unpruned_trees[t] = Iterate(WINDOW, INCREMENT);
    printf("\n");
    print_tree(unpruned_trees[t]);
    
    
    pruned_trees[t] = CopyTree(unpruned_trees[t]);
    if ( Prune(pruned_trees[t]) ) {
      printf("\nSimplified ");
      print_tree(pruned_trees[t]);
    }
    
    if ( pruned_trees[t]->Errors < pruned_trees[Best]->Errors ) 
      Best = t;
    
  }
  save_unpruned_trees(unpruned_trees);
  save_pruned_trees(pruned_trees);
  printf("\n--------\n");
  printf("\nBest tree from trial %d saved\n", Best);  
  return Best;
}



/*************************************************************************/
/*									 */
/*  The windowing approach seems to work best when the class		 */
/*  distribution of the initial window is as close to uniform as	 */
/*  possible.  FormTarget generates this initial target distribution,	 */
/*  setting up a TargetClassFreq value for each class.			 */
/*									 */
/*************************************************************************/
void FormTarget(int Size)
{
  int i, *ClassFreq;
  int c, Smallest, ClassesLeft=0;
  
  ClassFreq = (int *) calloc(MaxClass+1, sizeof(int));
  
  /*  Generate the class frequency distribution  */
  for (i=0; i <= MaxItem; i++) 
    ClassFreq[ Class(Item[i]) ]++;
  
  /*  Calculate the no. of classes of which there are items  */
  for (c=0; c <= MaxClass; c++) {
    if ( ClassFreq[c] ) {
      ClassesLeft++;
    } else {
      TargetClassFreq[c] = 0;
    }
  }

  while ( ClassesLeft ) {
    /*  Find least common class of which there are some items  */
    Smallest = -1;
    for (c=0; c <= MaxClass; c++) {
      if ( ClassFreq[c] &&
	   ( Smallest < 0 || ClassFreq[c] < ClassFreq[Smallest] ) ) {
	Smallest = c;
      }
    }

    /*  Allocate the no. of items of this class to use in the window  */
    TargetClassFreq[Smallest] = Min(ClassFreq[Smallest], Round(Size/ClassesLeft));
    ClassFreq[Smallest] = 0;
    Size -= TargetClassFreq[Smallest];
    ClassesLeft--;
  }

  free(ClassFreq);

  return;
}



/*************************************************************************/
/*									 */
/*  Form initial window, attempting to obtain the target class profile	 */
/*  in TargetClassFreq.  This is done by placing the targeted number     */
/*  of items of each class at the beginning of the set of data items.	 */
/*									 */
/*************************************************************************/
void FormInitialWindow()
{
  int i, Start=0, More;
  int c;
  
  shuffle();
  
  for (c=0; c <= MaxClass; c++) {
    More = TargetClassFreq[c];
    
    for ( i = Start ; More ; i++ ) {
      if ( Class(Item[i]) == c ) {
	Swap(Start, i);
	Start++;
	More--;
      }
    }
  }

  return;
}




/*************************************************************************/
/*									 */
/*  Grow a tree iteratively with initial window size Window and		 */
/*  initial window increment IncExceptions.				 */
/*									 */
/*  Construct a classifier tree using the data items in the		 */
/*  window, then test for the successful classification of other	 */
/*  data items by this tree.  If there are misclassified items,		 */
/*  put them immediately after the items in the window, increase	 */
/*  the size of the window and build another classifier tree, and	 */
/*  so on until we have a tree which successfully classifies all	 */
/*  of the test items or no improvement is apparent.			 */
/*									 */
/*  On completion, return the tree which produced the least errors.	 */
/*									 */
/*************************************************************************/
Tree Iterate(int Window, int IncExceptions)
{
  Tree Classifier, BestClassifier=NULL;
  int i, Errors, TotalErrors, BestTotalErrors=MaxItem+1,
    Exceptions, Additions;
  int Assigned;
  short Cycle=0;

  printf("Cycle   Tree    -----Cases----");
  printf("    -----------------Errors-----------------\n");
  printf("        size    window   other");
  printf("    window  rate   other  rate   total  rate\n");
  printf("-----   ----    ------  ------");
  printf("    ------  ----  ------  ----  ------  ----\n");
  
  do {
    /*  Build a classifier tree with the first Window items  */
    InitialiseWeights();
    AllKnown = True;
    Classifier = FormTree(0, Window,0);
    
    /*  Error analysis  */
    Errors = Round(Classifier->Errors);
    
    /*  Move all items that are incorrectly classified by the
	classifier tree to immediately after the items in the
	current window.  */
    Exceptions = Window;
    for (i=Window; i <= MaxItem; i++) {
      Assigned = Category(Item[i], Classifier);
      if ( Assigned != Class(Item[i]) ) {
	Swap(Exceptions, i);
	Exceptions++;
      }
    }
    Exceptions -= Window;
    TotalErrors = Errors + Exceptions;
    
    /*  Print error analysis  */
    printf("%3d  %7d  %8d  %6d  %8d%5.1f%%  %6d%5.1f%%  %6d%5.1f%%\n",
	   ++Cycle, TreeSize(Classifier), Window, MaxItem-Window+1,
	   Errors, 100*(float)Errors/Window,
	   Exceptions, 100*Exceptions/(MaxItem-Window+1.001),
	   TotalErrors, 100*TotalErrors/(MaxItem+1.0));
    
    /*  Keep track of the most successful classifier tree so far  */
    if ( ! BestClassifier || TotalErrors < BestTotalErrors ) {
      if ( BestClassifier ) ReleaseTree(BestClassifier);
      BestClassifier = Classifier;
      BestTotalErrors = TotalErrors;
    } else {
      ReleaseTree(Classifier);
    }

    /*  Increment window size  */
    Additions = Min(Exceptions, IncExceptions);
    Window = Min(Window + Max(Additions, Exceptions / 2), MaxItem + 1);
  } while ( Exceptions );
  
  return BestClassifier;
}
