#ifndef __UTILS_H_
#define __UTILS_H_
/* 
   utils.h
   $Id: utils.h,v 1.14 2005/03/15 20:48:40 eschris Exp $
	
	 Steven Eschrich
	 eschrich@csee.usf.edu
	 
   This file almost exclusively requires the use of GNU C extensions
   and GNU CPP extensions, although there is a warning if you do not
   use them. Don't expect any of these macros to work if you don't use
   gcc.


*/

#include "endian_config.h"


#define __UTILS_VERSION  "$Id: utils.h,v 1.14 2005/03/15 20:48:40 eschris Exp $"

/* System defines needed somewhere below */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <errno.h>
void exit();
extern int errno;

/********************************************************
 * Constants defined within libutils.
 * 
 ********************************************************/
/* The buffer size generally to create statically, a large number */
#define MAXBUF        10*1024
/* If arrays grow, by what increment */
#define REALLOC_INCR (1024)

/********************************************************
 * Data types defined within libutils.
 * 
 ********************************************************/
/* A boolean data type */
typedef int Boolean;
/* The boolean can take True/False values */
enum {False, True};
/* A plain, old byte */
typedef unsigned char BYTE;


/********************************************************
 * Variables defined within the library.
 * 
 ********************************************************/
/* An indicator of the endian-ness of the system */
extern int endian;

/* Constant bit-masks for simplicity. */
extern BYTE setbits[8];


/*****************************************************************************
   Macros
 
 	 Various useful macros (die, warn, info, etc). These macros ultimately
 	 call implementations after doing some massaging.
 	 
   Note: Variable length argument macros are available only to GNU C, as 
   are various aspects that make macros useful as pseudo-functions. This 
   functionality is not yet completely duplicated in ANSI C calls. 
*****************************************************************************/

/* Sensible string comparison */
#define STREQ(a,b)   ( !strcasecmp(a,b) )

/* Many logging utilities are defined separately */
#include "util_log.h"

/* These will only work with GNU C, a warning will be produced otherwise */
#ifdef __GNUC__

 /* MALLOC space, exiting the program on failure */
 #define MALLOC(size) \
         ({ \
            register void *__p; \
            if ( (__p=(void *)malloc(size)) == 0 ) { \
				      die_impl("%s:%d error allocating memory\n", \
                                                       __FILE__, __LINE__); \
            } \
            __p; \
	 })
	 /* CALLOC space, exiting program on failure */
  #define CALLOC(n,size) \
         ({ \
            register void *__p; \
            if ( (__p=(void *)calloc(n,size)) == 0 ) { \
	      			die_impl("%s:%d error allocating memory\n", \
                                          __FILE__, __LINE__); \
            } \
            __p; \
	 })
  /* REALLOC space, exiting program on failure */
  #define REALLOC(q, n) \
         ({ \
            register void *__p; \
            if ( (__p=(void *)realloc(q,n)) == 0 ) { \
				      die_impl("%s:%d error re-allocating memory\n", \
                                  				__FILE__, __LINE__); \
            } \
            __p;  \
	 })

  /* Open a file, exiting program on failure */
  #define FOPEN(filename, mode) \
       ({ \
          FILE *__fp; \
          char *__fstr=filename, *__mstr=mode; \
          extern int errno; \
          if ( (__fp=fopen(__fstr,__mstr)) == 0 ) { \
            die_impl("%s:%d\n Error opening %s (mode %s): %s\n", \
	                                     __FILE__, __LINE__, \
		                             __fstr, __mstr,strerror(errno)); \
          } \
          __fp; \
       })

  /* Open a file, exiting program on failure */
  #define GZOPEN(filename, mode) \
       ({ \
          gzFile __fp; \
          char *__fstr=filename, *__mstr=mode; \
          extern int errno; \
          if ( (__fp=gzopen(__fstr,__mstr)) == 0 ) { \
            die_impl("%s:%d\n Error opening %s (mode %s): %s\n", \
                                             __FILE__, __LINE__, \
                                             __fstr, __mstr,strerror(errno)); \
          } \
          __fp; \
       })

 /* The min and max macros are from the GNU cpp info pages */
 #define min(X, Y)                     \
     ({ typeof (X) __x = (X), __y = (Y);   \
        (__x < __y) ? __x : __y; })

 #define max(X, Y)                     \
     ({ typeof (X) __x = (X), __y = (Y);   \
        (__x > __y) ? __x : __y; })

 /* A perlism - chomp the last character off of the string */
 #define chomp(q) \
      do { \
         register char *__p=q; \
         register int __plen=strlen(__p); \
         if ( __p != 0 && __plen > 0 ) \
	   __p[__plen-1]=0; \
      } while(0) 

 /* A perlism, push a new element on the given array */
 #define push(array, element, array_count) \
      do { \
         register int __tmp_count=array_count; \
         if ( __tmp_count % REALLOC_INCR == 0 ) \
             array=(typeof(array))realloc(array, \
                                            (__tmp_count+REALLOC_INCR)* \
                                              sizeof(typeof(*array))); \
         if ( ! array ) { \
						die_impl("%s:%d Memory allocation failure\n",__FILE__,__LINE__);\
         }\
         array[__tmp_count]=element; \
      } while (0)
 /* A perlism, pop the last element off of the array */
 #define pop(array,array_count) \
        ({ \
          register int __tmp_count=array_count; \
          typeof(*array) __retval=0; \
          if ( __tmp_count-1 >= 0 ) \
             __retval=array[__tmp_count-1]; \
          __retval; \
        }) 

 

#else
 #error "Macros are implemented as functions, not sure it will work."
 FILE *FOPEN();
 void *MALLOC();
 void *CALLOC();
 void *REALLOC();
 void chomp();

 void die();
 void warn();
 void debug();
 void info();
/* The min and max macros are from the GNU cpp info pages */
 #ifndef min 
  #define min(X, Y)  ((X < Y)?X:Y)                    
  #warn "min defined unsafely"
 #else
   #error "min previously defined, not re-defined"
 #endif

 #ifndef max 
   #define max(X, Y) ((X > Y)?X:Y)                    \
   #warn "max defined unsafely"
 #else
   #error "max previously defined, not re-defined"
 #endif

#endif

/*****************************************************************************
  SECTION2: Inline functions

  Some code should be inlined if possible (particularly with gcc). However,
  if you don't compile with -O then the real versions of these functions
  must be used. So, a special script generates the below inline functions.
*****************************************************************************/


/* numbytes(n) - Given n bits, return number of bytes needed/used */
extern inline int numbytes(int n)
{
   register unsigned int i = n;
   register unsigned int d = i>>3;

   if ( (i&0x7) > 0 ) d++;

   return d;
}

/* bget(v,index) - Return the index'th bit in the bitstring v */
extern inline int bget(BYTE *v,int index)
{
   register unsigned int i = index;
   register unsigned int o = i & 0x7;
   register unsigned int d = i >> 3;

   return ((v[d] & setbits[o]) > 0);
}


/* bset(v,index,value) - Set the index'th bit in the bitstring v to value */
extern inline void bset(BYTE *v,int index,int value)
{
   register unsigned int i = index;
   register unsigned int o = i & 0x07;
   register unsigned int d = i >> 3;
   if ( (value) == 0 )
      v[d] &= (BYTE)~setbits[o];
   else
      v[d] |= setbits[o];

   return;
}

/* Remove comments from line, given comment char c */
extern inline int strip_comments(char *p, char c)
{
   if (!p) return -1;
   for (;*p!=0; p++) {
      if (*p==c) {
         *p=0;
          break;
      }
   }
   return 0;
}

/* Trim leading spaces (whitespace) */
extern inline char *ltrim(char *start)
{
  if ( !start ) return start;
  for (; *start != 0 && isspace((int)(*start)); start++);
  return start;
}

/* Seek from the end of p backwards to the first non-blank character.
   This is a bit difficult since array bounds must be kept in mind. */
extern inline char *rtrim(char *start, char *end)
{
    if ( !start ) return 0;
  /* Seek back to beginning or first non-blank */
  for (; end >start && (*end == 0 || isspace((int)(*end))); end--);
  /* Are we at the beginning? If so, we might either have a blank
     or a non-blank. */
  if ( end == start ) {
    if ( isspace((int)(*end)) )  *end=0;         /* \n or the like */
    else if ( *end != 0 ) *(end+1)=0;   /* non-empty string, zero next char */
  } else                *(end+1)=0;
  return end;
}

extern inline int endsWith(char *str, char *sub)
{
  int N=strlen(str);
  int last=strlen(sub);

  if ( last > N ) return 0;
  
  return (!strcasecmp(str+N-last, sub) );
}


/*****************************************************************************
   SECTION3: Regular functions

   Here are functions that are normally implemented within libutils.a for
   general use. There is no need to inline, etc.
*****************************************************************************/
char *rtrim();
char *ltrim();
char *trim();
char **split();
/* Compare two doubles (for sorting) */
int dcompare(const void *p1, const void *p2);
int icompare(const void *p1, const void *p2);
int fcompare(const void *p1, const void *p2);



/* Matrix allocation */
double **create_matrix(int rows, int columns);
void     free_matrix(double **m);

int swap_int(unsigned int i);
short swap_short(unsigned short i);


#endif
